@@ -8,8 +8,8 @@ Classes
 -------
 * :class:`Atom` : Check the inputs from the :class:`atoMEC.Atom` object.
 * :class:`ISModel` : Check the inputs from the :obj:`atoMEC.models.ISModel` class.
-* :class:`EnergyCalcs` : Check the inputs from the\
- :func:`atoMEC.models.ISModel.CalcEnergy` function.
+* :class:`EnergyCalcs` : Check the inputs from the \
+                         :func:`atoMEC.models.ISModel.CalcEnergy` function.
 * :class:`InputError` : Exit atoMEC and print relevant input error message.
 * :class:`InputWarning` : Warn if inputs are considered outside of typical ranges.
 """
@@ -261,9 +261,9 @@ class Atom:
         if not isinstance(density, (float, intc)):
             raise InputError.density_error("Density is not a number")
         else:
-            if density > 100 or density < 0:
+            if density > 1000 or density < 0:
                 raise InputError.density_error(
-                    "Density must be a positive number less than 100"
+                    "Density must be a positive number less than 1000"
                 )
 
         return density
@@ -313,9 +313,10 @@ class Atom:
                 else:
                     density = self.radius_to_dens(atom, radius)
             elif radius == -1 and density != -1:
-                if density > 100 or density < 0:
+                max_density = 1000
+                if density > max_density or density < 0:
                     raise InputError.density_error(
-                        "Density must be a positive number less than 100"
+                        f"Density must be a positive number less than {max_density}"
                     )
                 else:
                     radius = self.dens_to_radius(atom, density)
@@ -415,9 +416,9 @@ class ISModel:
             by the current version of atoMEC
         """
         # supported families of libxc functional by name
-        names_supp = ["lda"]
+        names_supp = ["lda", "gga"]
         # supported families of libxc functional by id
-        id_supp = [1]
+        id_supp = [1, 2]
 
         # check both the exchange and correlation functionals are valid
         xc_func, err_xc = xc.check_xc_func(xc_func, id_supp)
@@ -443,7 +444,7 @@ class ISModel:
 
         return xc_func
 
-    def check_unbound(unbound):
+    def check_unbound(unbound, bc):
         """
         Check the unbound electron input is accepted.
 
@@ -451,6 +452,8 @@ class ISModel:
         ----------
         unbound : str
             defines the treatment of the unbound electrons
+        bc : str
+            the boundary condition
 
         Returns
         -------
@@ -463,22 +466,27 @@ class ISModel:
             if the treatment of unbound electrons is not a valid input
         """
         # list all possible treatments for unbound electrons
-        unbound_permitted = ["ideal"]
-
-        # convert unbound to all lowercase
-        unbound.lower()
+        unbound_permitted = ["ideal", "quantum"]
 
         if not isinstance(unbound, str):
             raise InputError.unbound_error(
                 "Unbound electron description is not a string"
             )
         else:
+            # convert unbound to all lowercase
+            unbound.lower()
+
             if unbound not in unbound_permitted:
                 err_msg = (
                     "Treatment of unbound electrons not recognised. \n                "
-                    " Allowed treatments are: " + [ub for ub in unbound_permitted]
+                    " Allowed treatments are: "
+                    + ", ".join([ub for ub in unbound_permitted])
                 )
                 raise InputError.unbound_error(err_msg)
+            elif bc == "bands" and unbound == "ideal":
+                err_msg = "bands bc cannot be used with ideal treatment of unbound \
+electrons. Only quantum treatment is permitted."
+                raise InputError.unbound_error(err_msg)
 
         return unbound
 
@@ -503,18 +511,18 @@ class ISModel:
             if the boundary condition is not recognised
         """
         # list permitted boundary conditions
-        bcs_permitted = ["dirichlet", "neumann"]
+        bcs_permitted = ["dirichlet", "neumann", "bands"]
 
-        # convert to lowercase
-        bc.lower()
-
-        if not isinstance("bc", str):
+        if not isinstance(bc, str):
             raise InputError.bc_error("Boundary condition is not a string")
         else:
+            # convert to lowercase
+            bc.lower()
+
             if bc not in bcs_permitted:
                 err_msg = (
                     "Boundary condition is not recognised. \n                 Allowed"
-                    " boundary conditions are: " + [b for b in bcs_permitted]
+                    " boundary conditions are: " + ", ".join([b for b in bcs_permitted])
                 )
                 raise InputError.bc_error(err_msg)
 
@@ -666,9 +674,11 @@ class EnergyCalcs:
         grid_params : dict
             dictionary of grid parameters as follows:
             {
-            `ngrid` (``int``)   : number of grid points,
-            `x0`    (``float``) : LHS grid point takes form
+            `ngrid` (``int``)        : number of grid points,
+            `x0`    (``float``)      : LHS grid point takes form
             :math:`r_0=\exp(x_0)`; :math:`x_0` can be specified
+            `ngrid_coarse` (``int``) : (smaller) number of grid points for estimation
+            of eigenvalues with full diagonalization
             }
 
         Raises
@@ -678,7 +688,7 @@ class EnergyCalcs:
         InputError.ngrid_warning
             if `ngrid` is outside a reasonable convergence range
         """
-        # First assign the keys ngrid and x0 if they are not given
+        # First assign the keys ngrid, x0 and ngrid_coarse if they are not given
         try:
             ngrid = grid_params["ngrid"]
         except KeyError:
@@ -689,6 +699,11 @@ class EnergyCalcs:
         except KeyError:
             x0 = config.grid_params["x0"]
 
+        try:
+            ngrid_coarse = grid_params["ngrid_coarse"]
+        except KeyError:
+            ngrid_coarse = config.grid_params["ngrid_coarse"]
+
         # check that ngrid is an integer
         if not isinstance(ngrid, intc):
             raise InputError.grid_error("Number of grid points not an integer!")
@@ -700,13 +715,24 @@ class EnergyCalcs:
         elif ngrid > 5000:
             print(InputWarning.ngrid_warning("high", "expensive"))
 
+        # check that ngrid_coarse is an integer
+        if not isinstance(ngrid_coarse, intc):
+            raise InputError.grid_error("Number of coarse grid points not an integer!")
+        # check that ngrid_coarse is a positive number
+        if ngrid_coarse < 0:
+            raise InputError.grid_error("Number of coarse grid points must be positive")
+        elif ngrid_coarse < 100:
+            print(InputWarning.ngrid_warning("low", "inaccurate"))
+        elif ngrid_coarse > 500:
+            print(InputWarning.ngrid_warning("high", "expensive"))
+
         # check that x0 is reasonable
         if x0 > -3:
             raise InputError.grid_error(
                 "x0 is too high, calculation will likely not converge"
             )
 
-        grid_params = {"ngrid": ngrid, "x0": x0}
+        grid_params = {"ngrid": ngrid, "x0": x0, "ngrid_coarse": ngrid_coarse}
 
         return grid_params
 
@@ -726,9 +752,10 @@ class EnergyCalcs:
         conv_params : dict of floats
             dictionary of convergence parameters as follows:
             {
-            `econv` (``float``) : convergence for total energy,
-            `nconv` (``float``) : convergence for density,
-            `vconv` (``float``) : convergence for electron number
+            `econv` (``float``)  : convergence for total energy,
+            `nconv` (``float``)  : convergence for density,
+            `vconv` (``float``)  : convergence for electron number,
+            `eigtol` (``float``) : tolerance for eigenvalues
             }
 
         Raises
@@ -738,7 +765,7 @@ class EnergyCalcs:
         """
         conv_params = {}
         # loop through the convergence parameters
-        for conv in ["econv", "nconv", "vconv"]:
+        for conv in ["econv", "nconv", "vconv", "eigtol"]:
             # assign value if not given
             try:
                 x_conv = input_params[conv]
@@ -808,6 +835,61 @@ class EnergyCalcs:
 
         return scf_params
 
+    @staticmethod
+    def check_band_params(input_params):
+        r"""
+        Check if band parameters are reasonable, or assign if empty.
+
+        Parameters
+        ----------
+        input_params : dict
+            can contain the keys `maxscf` and `mixfrac` for max scf cycle
+            and potential mixing fraction
+
+        Returns
+        -------
+        band_params : dict
+            dictionary for band parameters as follows:
+            {
+            `nkpts`   (``int``)   : number of levels per band,
+            `de_min`   (``float``) : minimum energy gap to make a band
+            }
+
+        Raises
+        ------
+        InputError.bands_error
+            if band parameters are of invalid type or range
+        """
+        band_params = {}
+
+        for p in ["nkpts", "de_min"]:
+            band_params[p] = input_params.get(p, config.band_params[p])
+
+        # dirichlet and neumann bcs should only have one band
+        bcs_no_bands = ["dirichlet", "neumann"]
+        if config.bc in bcs_no_bands:
+            band_params["nkpts"] = 1
+
+        # check the number of bands is valid
+        else:
+            if not isinstance(band_params["nkpts"], intc):
+                raise InputError.bands_error("nkpts is not an integer")
+            else:
+                min_k_points = 10
+                if band_params["nkpts"] < min_k_points:
+                    raise InputError.bands_error(
+                        f"bands requires at least {min_k_points} k points"
+                    )
+
+        # check the minimum band spacing is valid
+        if not isinstance(band_params["de_min"], (float, intc)):
+            raise InputError.bands_error("de_min is not a number")
+        else:
+            if band_params["de_min"] < 0:
+                raise InputError.bands_error("de_min must be positive")
+
+        return band_params
+
 
 class InputError(Exception):
     """Exit atoMEC and print relevant input error message."""
@@ -941,6 +1023,22 @@ class InputError(Exception):
         print("Error in boundary condition input: " + err_msg)
         sys.exit("Exiting atoMEC")
 
+    def bands_error(err_msg):
+        """
+        Raise exception if `nkpts` not positive int or `dE_spc` not positive number.
+
+        Parameters
+        ----------
+        err_msg : str
+            the error message printed
+
+        Returns
+        -------
+        None
+        """
+        print("Error in bands input: " + err_msg)
+        sys.exit("Exiting atoMEC")
+
     def spinpol_error(err_msg):
         """
         Raise exception if spinpol not a boolean.
@@ -1005,6 +1103,22 @@ class InputError(Exception):
         print("Error in scf_params input: " + err_msg)
         sys.exit("Exiting atoMEC")
 
+    def v_shift_error(err_msg):
+        """
+        Raise exception if error in SCF inputs.
+
+        Parameters
+        ----------
+        err_msg : str
+            the error message printed
+
+        Returns
+        -------
+        None
+        """
+        print("Error in v_shift input: " + err_msg)
+        sys.exit("Exiting atoMEC")
+
 
 class InputWarning:
     """Warn if inputs are considered outside of typical ranges, but proceed anyway."""
